/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/
/**
 * @file src/applications/afs/gtkui/saveas.c
 * @author Christian Grothoff
 * @author Igor Wronsky
 *
 * This code opens a save-as window.
 **/

#include "config.h"
#include <string.h>
#include <gtk/gtk.h>

#include "util/logging.h"
#include "util/xmalloc.h"
#include "applications/afs/gtkui/download.h"
#include "applications/afs/gtkui/helper.h"

typedef struct {
  RootNode root;
  GtkWidget * w;
} SaveAs;

/**
 * Destroy the DownloadModel data structure of the
 * saveas dialog.
 *
 * @param dlm the DownloadModel to be destroyed
 **/
static gint destroySaveAs(GtkWidget * widget,
			  SaveAs * saveas) {
  LOG(LOG_DEBUG, 
      "DEBUG: destroying saveas window (%x)\n", 
      saveas);
  FREE(saveas);
  return TRUE;
}
 
/**
 * Get the selected filename and start downloading 
 *
 * @param root data required for the download
 **/
static gint file_ok_sel(GtkWidget * okButton,
			SaveAs * saveas) {
  gchar * filename;

  LOG(LOG_DEBUG, 
      "DEBUG: file_ok_sel (%x)\n", 
      saveas);
  filename
    = gtk_file_selection_get_filename(GTK_FILE_SELECTION(saveas->w));
  startDownload(filename, &saveas->root);
  gtk_widget_destroy(saveas->w);
  /* destroySaveAs does: "FREE(saveas);" */
  return FALSE;
}

/**
 * Open the window that prompts the user for the 
 * filename.
 * This method must open the window,
 * copy the arguments and return. After the method
 * returns, the arguments passed to it will be
 * freed, so pointer should not be retained.
 * The method executes during a signal handler,
 * so a GTK lock is not required to to GUI 
 * operations.
 * 
 * @param dlm Data required for the download
 **/
void openSaveAs(RootNode * root) {
  GtkWidget * window;
  SaveAs * saveas;
  
  saveas = MALLOC(sizeof(SaveAs));
  memcpy(&saveas->root,
	 root,
	 sizeof(RootNode));

  window = gtk_file_selection_new("save as");
  saveas->w = window;

  /* if the search result specified a suggested
     filename, fill it in! */
  if (root->header.filename != NULL)
    gtk_file_selection_set_filename(GTK_FILE_SELECTION(window),
				    root->header.filename);

  /* callbacks (destroy, ok, cancel) */
  gtk_signal_connect(GTK_OBJECT(window),
		     "destroy",
		     GTK_SIGNAL_FUNC(destroySaveAs),
		     saveas);
  /* Connect the ok_button to file_ok_sel function */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(window)->ok_button),
		     "clicked", 
		     GTK_SIGNAL_FUNC(file_ok_sel), 
		     saveas);
  /* Connect the cancel_button to destroy the widget */
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(window)->cancel_button),
		     "clicked", 
		     GTK_SIGNAL_FUNC(destroyWidget), 
		     window);
  gtk_widget_show(window);
}

/* end of saveas.c */
