/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file src/applications/afs/gtkui/search.c
 * @author Christian Grothoff
 * @author Igor Wronsky
 *
 * This handles the box displaying search results for
 * the gtk+ client.
 **/

#include "config.h"

#include <errno.h>
#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "util/configuration.h"
#include "util/logging.h"
#include "util/returnvalues.h"
#include "util/xmalloc.h"
#include "applications/afs/gtkui/helper.h"
#include "applications/afs/gtkui/download.h"
#include "applications/afs/gtkui/saveas.h"
#include "util/port.h"

/**
 * This method is called whenever the user clicks the
 * download button.  It opens the "save-as" dialog.
 *
 * @param listModel Data related to the search
 **/
static void download(GtkWidget * widget,
		     ListModel * listModel) {  
  RootNode * root;
  int descLen;
  int nameLen;
  gchar * tmp;
  gint row;
  
  row = listModel->selectedRow;
  if (row == -1) {
    guiMessage("Nothing selected!\n");
    return;
  }

  root = MALLOC(sizeof(RootNode));
  memset(root, 
  	 0,
	 sizeof(RootNode));

  /* BEGIN reconstruction root node */
  /* Collect data required from clist entry 
     (reconstruct root node; FIXME: add mime-type) */
  gtk_clist_get_text(GTK_CLIST(listModel->search_result_list),
 		     row,
		     1,
		     &tmp);
  sscanf(tmp, 
	 "%d",
	 &root->header.fileIdentifier.file_length);
  gtk_clist_get_text(GTK_CLIST(listModel->search_result_list),
 		     row,
		     3,
		     &tmp);
  sscanf(tmp,
	 "%d",
	 &root->header.fileIdentifier.crc);
  gtk_clist_get_text(GTK_CLIST(listModel->search_result_list),
 		     row,
		     4,
		     &tmp);
  hex2hash((HexName*)tmp,
  	   &root->header.fileIdentifier.chk.query);
  gtk_clist_get_text(GTK_CLIST(listModel->search_result_list),
 		     row,
		     5, 
		     &tmp);
  hex2hash((HexName*)tmp,
  	   &root->header.fileIdentifier.chk.key);
  gtk_clist_get_text(GTK_CLIST(listModel->search_result_list),
 		     row, 
		     0,
		     &tmp);
  descLen = strlen(tmp);
  if(descLen >= MAX_DESC_LEN) {
    LOG(LOG_ERROR,
        "ERROR: ASSERTION failed: description too long? (%s)",
	tmp);
    descLen = MAX_DESC_LEN-1;
  }
  memcpy(&root->header.description,
  	 tmp,
	 descLen+1);
  root->header.description[descLen] = 0;
  gtk_clist_get_text(GTK_CLIST(listModel->search_result_list),
 		     row, 
		     2, 
		     &tmp);
  nameLen = strlen(tmp);
  if (nameLen >= MAX_FILENAME_LEN) {
    LOG(LOG_ERROR,
        "ERROR: ASSERTION failed: filename too long? (%s)",
	tmp);
    nameLen = MAX_FILENAME_LEN-1;
  }
  memcpy(root->header.filename,
  	 tmp,
	 nameLen);
  root->header.filename[nameLen] = 0;
  /* END of reconstruction of root node */

  /* Remove entry from search results.  Yes,
     if the user cancel's the download, the
     entry does not re-appear.  That's intended,
     after all, if you cancel, it's probably because
     it took too long to download anyway... 
     If you really need it back, just search again! */
  gtk_clist_freeze(GTK_CLIST(listModel->search_result_list));
  gtk_clist_remove(GTK_CLIST(listModel->search_result_list),
  		   listModel->selectedRow);
  gtk_clist_thaw(GTK_CLIST(listModel->search_result_list));
  listModel->selectedRow = -1;

  /* open the "save-as" window */ 
  openSaveAs(root);
  FREE(root);
}

/**
 * User has selected some search list row
 **/
static void select_row_callback(GtkWidget *widget,
				gint row,
				gint column,
				GdkEventButton *event,
				gpointer data) {
  if(GTK_CLIST(((ListModel*)data)->search_result_list)->freeze_count>0)
    return;
  ((ListModel*)data)->selectedRow = row;
}

/**
 * Some search list row has been unselected
 **/
static void unselect_row_callback(GtkWidget *widget,
				  gint row,
				  gint column,
				  GdkEventButton *event,
				  gpointer data) {
  if(GTK_CLIST(((ListModel*)data)->search_result_list)->freeze_count>0)
    return;
  ((ListModel*)data)->selectedRow = -1;
}

/**
 * Display results.  This is a callback from receiveResults that is
 * called on every new result.
 *
 * @param rootNode Data about a file
 * @param model Data related to the search
 **/
static void displayResult(RootNode * rootNode,
			  ListModel * model) {
  GtkWidget * search_result_list;
  gchar * results[7];
  int i;

  /* ensure well-formed strings */
  rootNode->header.description[MAX_DESC_LEN-1] = 0;
  rootNode->header.filename[MAX_FILENAME_LEN-1] = 0;
  rootNode->header.mimetype[MAX_MIMETYPE_LEN-1] = 0;

  search_result_list = model->search_result_list;
  results[0] = STRDUP(rootNode->header.description);
  results[1] = MALLOC(32);
  sprintf(results[1],"%d", 
	  ntohl(rootNode->header.fileIdentifier.file_length));
  results[2] = STRDUP(rootNode->header.filename);
  results[3] = MALLOC(32);
  sprintf(results[3],"%d", 
	  ntohl(rootNode->header.fileIdentifier.crc));
  results[4] = MALLOC(sizeof(HexName));
  hash2hex(&rootNode->header.fileIdentifier.chk.query,
	   (HexName*) results[4]);
  results[5] = MALLOC(sizeof(HexName));
  hash2hex(&rootNode->header.fileIdentifier.chk.key,
           (HexName*) results[5]);
  results[6] = STRDUP(rootNode->header.mimetype);
  
  gdk_threads_enter();
  gtk_clist_freeze(GTK_CLIST(search_result_list));
  gtk_clist_append(GTK_CLIST(search_result_list),
		   results);
  gtk_clist_thaw(GTK_CLIST(search_result_list));
  gdk_threads_leave();

  for (i=0;i<7;i++)
    FREE(results[i]);
}

/**
 * Struct to pass a couple of arguments to a new
 * thread "receiveResults_".
 **/
typedef struct {
  int num_keywords;
  HashCode160 * keys;
  AFS_CS_QUERY ** messages;
  void * displayResult;
  ListModel * model;
} _receiveResultArgs_;

/**
 * Should the receive thread abort? (has the window been closed?)
 * This is a callback for the receiveResults method (since
 * not every "error" on the socket corresponds to a closed
 * window!).
 *
 * @return YES if we should abort
 **/
static int testTermination(ListModel * model) {
  if (model->sem == NULL)
    return NO;
  else
    return YES;
}

/**
 * Main method of the receiveResults-threads.
 * Runs the receiveResults method and frees some
 * data structures once the search is aborted.
 * See also "stopSearch".
 **/
static void receiveResults_(_receiveResultArgs_ * args) {
  receiveResults(args->model->SEARCH_socket_,	
  		 args->num_keywords,
		 args->keys,
		 args->messages,
		 args->displayResult,
		 args->model,
		 (TestTerminateThread)&testTermination,
		 args->model);
  FREE(args->keys);
  SEMAPHORE_UP(args->model->sem); /* signal: thread terminated */
  FREE(args);
}

/**
 * The main method of the search-thread.
 *
 * @param searchString What to look for
 * @param model Data related to the search
 * @return OK on success, SYSERR on error
 **/
static int startSearchThread(char * searchString,
			      ListModel * model) {
  HashCode160 * keys;
  int num_keywords;
  AFS_CS_QUERY ** messages;
  _receiveResultArgs_ * receiveArgs;
  SendQueriesContext * sqc;
  pthread_t receiveThread;

  num_keywords = parseKeywords(1,
			       &searchString,
			       &keys);
  if (num_keywords <= 0) 
    return SYSERR;
  
  buildMessages(num_keywords,
		keys,
		&messages);
  sqc = MALLOC(sizeof(SendQueriesContext));
  sqc->messages = messages;
  sqc->queryCount = num_keywords;
  sqc->sock = model->SEARCH_socket_;
  sqc->timeout = getConfigurationInt("AFS",
				     "SEARCHTIMEOUT") * cronSECONDS; 
  cronTime(&sqc->start);
  model->sqc = sqc;
  addCronJob((CronJob)&sendQueries,
	     0, 
	     0,
	     sqc);

  receiveArgs = MALLOC(sizeof(_receiveResultArgs_));
  receiveArgs->num_keywords = num_keywords;
  receiveArgs->keys = keys;
  receiveArgs->messages = messages;
  receiveArgs->displayResult = displayResult;
  receiveArgs->model = model;  
  if (0 != PTHREAD_CREATE(&receiveThread,
			  (void * (*)(void *)) &receiveResults_,
			  receiveArgs,
			  16 * 1024)) {
    errexit("FATAL: could not create receive thread (%s)!\n",
	    strerror(errno));    
  } else
    pthread_detach(receiveThread);
  return OK;
}

/**
 * Stop the search thread and free the model.  This method MUST always
 * be called when the search ends, either because the search was
 * aborted or because gnunet-gtk exists as a whole.  
 *
 * @param widget the window (not used)
 * @param model the model with the socket and the sqc 
 **/
static void stopSearch(GtkWidget * widget,
		       ListModel * model) { 
  int i;
  LOG(LOG_DEBUG, 
      "DEBUG: stopSearch called\n");

  /* the terminated search thread ups this semaphore
     once it is done and we can free data structures */
  model->sem = SEMAPHORE_NEW(0);

  /* this signals the download thread to terminate */
  destroySocket(model->SEARCH_socket_); 

  /* stop the cron-job that does the requests */
  suspendCron(); 
  delCronJob((CronJob)&sendQueries, 
	     0, 
	     model->sqc);
  resumeCron();

  /* wait for download thread signal */
  SEMAPHORE_DOWN(model->sem);

  /* Now we can finally free the shared data structures.
     Note that the terminated search thread freed 
     some of the memory that was allocated in
     startSearchThread (see receiveResults_)
     we free the rest. */
  SEMAPHORE_FREE(model->sem);  
  releaseClientSocket(model->SEARCH_socket_);
  for (i=0;i<model->sqc->queryCount;i++) 
    FREE(model->sqc->messages[i]);  
  FREE(model->sqc->messages);
  FREE(model->sqc);
  FREE(model);
}

/**
 * A function for case-insensitive text comparisons
 **/
static gint string_casecmp(GtkCList *clist,
			   gconstpointer ptr1,
			   gconstpointer ptr2) {
  char * text1;
  char * text2;
  GtkCListRow * row1 = (GtkCListRow *) ptr1;
  GtkCListRow * row2 = (GtkCListRow *) ptr2;

  text1 = GTK_CELL_TEXT(row1->cell[clist->sort_column])->text;
  text2 = GTK_CELL_TEXT(row2->cell[clist->sort_column])->text;
 
  return (strcasecmp(text1,text2));	
}

/**
 * A function for numeric comparisons
 **/
static gint string_numericcmp(GtkCList *clist, 
			      gconstpointer ptr1,
			      gconstpointer ptr2) {
  int value1;
  int value2;
  GtkCListRow * row1 = (GtkCListRow *) ptr1;
  GtkCListRow * row2 = (GtkCListRow *) ptr2;
  
  value1 = atoi(GTK_CELL_TEXT(row1->cell[clist->sort_column])->text);
  value2 = atoi(GTK_CELL_TEXT(row2->cell[clist->sort_column])->text);
  
  if(value1>value2)
    return(-1);
  else if(value1==value2)
    return(0);
  else
    return(1);
}

/**
 * Changes the current sort column and sorts the list.
 **/
static void sort_column_callback(GtkCList * clist,
				 gint column,
				 gpointer data) {
  /* Sort all columns as strings except 1 (size) */
  if (column == 1)
    gtk_clist_set_compare_func(clist, 
			       (GtkCListCompareFunc)string_numericcmp);
  else
    gtk_clist_set_compare_func(clist, 
			       (GtkCListCompareFunc)string_casecmp);  
  gtk_clist_set_sort_column(clist, column);
  gtk_clist_freeze(clist);
  gtk_clist_sort(clist);
  gtk_clist_thaw(clist);
}

/** 
 * Returns a box containing the search results list.
 **/
GtkWidget * getSearchWindow(gchar * title) {
  GtkWidget * scrolled_window;
  GtkWidget * button;
  GtkWidget * box;
  GtkWidget * search_result_list;
  ListModel * model;  
  static gchar * descriptions[] = {
    "Description",
    "Size",
    "Filename",
    "CRC",
    "HASH1",
    "HASH2",
    "Mimetype"
  };
  /* widths of the colums in the search results */
  static int widths[] = {
    470, 70, 200, 40, 40, 40, 200,
  };
  int i;
  int ok;

  model = (ListModel*) MALLOC(sizeof(ListModel));
  model->selectedRow = -1; /* none selected */
  model->sem = NULL; /* semaphore used for shutdown sync */
  box = gtk_vbox_new(FALSE, 0);
  
  /* scrolled window for the list */
  scrolled_window = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
                                 GTK_POLICY_AUTOMATIC, 
				 GTK_POLICY_ALWAYS);
  gtk_box_pack_start(GTK_BOX(box), 
		     scrolled_window, 
		     TRUE,
		     TRUE, 
		     0);
  gtk_widget_show(scrolled_window);
  
  /* result set list */
  search_result_list 
    = gtk_clist_new_with_titles(7, descriptions);
  model->search_result_list 
    = search_result_list;
  model->SEARCH_socket_
    = getClientSocket();
  if (model->SEARCH_socket_ == NULL) {
    gtk_widget_destroy(box);
    FREE(model);
    return NULL;
  }

  gtk_container_add(GTK_CONTAINER(scrolled_window), 
		    search_result_list);
  gtk_widget_show(search_result_list);

  gtk_clist_set_selection_mode
    (GTK_CLIST(search_result_list),
     GTK_SELECTION_SINGLE);
  /* set passive titles as default */
  gtk_clist_column_titles_passive
    (GTK_CLIST(search_result_list));

  /* allow sorting by description, size, filename and mimetype */
  gtk_clist_column_title_active(GTK_CLIST(search_result_list),
				0);
  gtk_clist_column_title_active(GTK_CLIST(search_result_list),
				1);
  gtk_clist_column_title_active(GTK_CLIST(search_result_list),
				2);
  gtk_clist_column_title_active(GTK_CLIST(search_result_list),
				6);
  gtk_signal_connect(GTK_OBJECT(search_result_list),
		     "click-column",
		     GTK_SIGNAL_FUNC(sort_column_callback),
		     NULL);

  /* description left, size right justification */
  gtk_clist_set_column_justification
    (GTK_CLIST(search_result_list),
     0,
     GTK_JUSTIFY_LEFT);
  gtk_clist_set_column_justification
    (GTK_CLIST(search_result_list),
     1,
     GTK_JUSTIFY_RIGHT);
  gtk_signal_connect(GTK_OBJECT(search_result_list),
		     "unselect_row",
		     GTK_SIGNAL_FUNC(unselect_row_callback),
		     (gpointer) model);
  gtk_signal_connect(GTK_OBJECT(search_result_list),
		     "select_row",
		     GTK_SIGNAL_FUNC(select_row_callback),
		     (gpointer) model);

  /* set column widths */
  for (i=0;i<7;i++)
    gtk_clist_set_column_width(GTK_CLIST(search_result_list), 
			       i,
			       widths[i]);

  /* download button */
  button = gtk_button_new_with_label("Download");
  gtk_signal_connect (GTK_OBJECT(button), 
		      "clicked",
		      GTK_SIGNAL_FUNC(download), 
		      model);
  gtk_box_pack_start(GTK_BOX(box), 
		     button,
		     FALSE,
		     FALSE,
		     0);
  gtk_widget_show(button);

  /* generic: on delete request, just do it (always OK) */
  gtk_signal_connect(GTK_OBJECT(scrolled_window), 
                     "delete_event",
		     GTK_SIGNAL_FUNC(deleteEvent), 
		     NULL);
  /* when we are destroyed (e.g. search aborted), stop the search thread */
  gtk_signal_connect(GTK_OBJECT(scrolled_window), 
                     "destroy",
		     GTK_SIGNAL_FUNC(stopSearch), 
		     model); 

  /* start searching */
  ok = startSearchThread(title,
			 model);
  if (ok == SYSERR) {
    releaseClientSocket(model->SEARCH_socket_);
    gtk_widget_destroy(box);
    FREE(model);
    return NULL;
  } else {
    /* return complete box such that it gets displayed */
    return box;
  }
}

/* end of search.c */
